/*****************************************************************************
 * game.cpp - General game loop class
 *****************************************************************************/
#include <stdexcept>
#include "SDL.h"
#include "actioneventqueue.h"
#include "ccmap.h"
#include "common.h"
#include "config.h"
#include "cursor.h"
#include "dispatcher.h"
#include "game.h"
#include "graphicsengine.h"
#include "input.h"
#include "loadingscreen.h"
#include "logger.h"
#include "playerpool.h"
#include "sidebar.h"
#include "soundengine.h"
#include "vqa.h"
#include "ai.h"
#include "menu.h"
#include "vfs.h"
#include "mapsini.h"

using std::string;
using std::runtime_error;

/** Constructor, loads the map, sidebar and such. plays briefing and actionmovie
 */
Game::Game() {
	MissionNr			= 0;
	OldUptime			= 0;
	pc::Ai				= NULL;
	pc::input			= NULL;
	pc::cursor			= NULL;
	pc::sidebar 		= NULL;
	pc::MissionsMapdata	= NULL;
	p::dispatcher		= NULL;
	p::aequeue			= NULL;
	p::ccmap			= NULL;
	pc::imagepool		= NULL;
	pc::gfxeng			= NULL;
//	pc::sfxeng			= NULL;
	BattleControlTerminated = false;
}

/** Destructor, frees up some memory */
Game::~Game()
{
	FreeMemory();
}

/** Game menu (early game start) */
void Game::HandleGameMenu()
{
MenuClass Menu;

	printf ("%s line %i: Start display menu\n", __FILE__, __LINE__);

	// Start with a clean image cache
	pc::imgcache->Cleanup();

	// Draw the menu and handle menu input
	Menu.HandleMenu();

	// Cleanup the image cache again
	pc::imgcache->Cleanup();
}


/** Initialize game (after the menu) */
void Game::InitializeMap( std::string MapName )
{
	gamemode = pc::Config.gamemode;

	p::aequeue = new ActionEventQueue();

	try {
		p::ccmap = new CnCMap();
		p::ccmap->InitCnCMap();
		p::ccmap->loadMap( pc::Config.mapname.c_str(), NULL );
	} catch (CnCMap::LoadMapError&) {
		// loadmap will have printed the error
		throw GameError();
	}

    p::dispatcher = new Dispatcher::Dispatcher();
    switch (pc::Config.dispatch_mode) {
        case 0:
            break;
        case 1:
            // Record
            break;
        case 2:
            // Playback
            break;
        default:
            logger->error("Invalid dispatch mode: %i\n",pc::Config.dispatch_mode);
            throw GameError();
            break;
    }


    switch (gamemode) {
    case 0:
        try {
		printf ("%s line %i: Brief = %s\n", __FILE__, __LINE__, p::ccmap->getMissionData().brief);
            VQAMovie mov(p::ccmap->getMissionData().brief);
            mov.play();
        } catch (std::runtime_error&) {
        }
        try {
		printf ("%s line %i: Action = %s\n",  __FILE__, __LINE__, p::ccmap->getMissionData().action);
            VQAMovie mov(p::ccmap->getMissionData().action);
            mov.play();
        } catch (std::runtime_error&) {
        }
        break;
    case 1:
    case 2:
    default:
        break;
    }

    /* init sidebar */
    try {
        pc::sidebar = new Sidebar( p::ppool->getLPlayer(), pc::gfxeng->getHeight(), p::ccmap->getMissionData().theater );
    } catch (Sidebar::SidebarError&) {
        throw GameError();
    }

    /* init cursor */
    pc::cursor = new Cursor();

    /* init the input functions */
    pc::input = new Input( pc::gfxeng->getWidth(), pc::gfxeng->getHeight(), pc::gfxeng->getMapArea() );
}

void Game::FreeMemory()
{
	if (pc::imgcache != NULL)
		pc::imgcache->Cleanup();
	if (pc::input != NULL)
		delete pc::input;
	if (pc::cursor != NULL)
		delete pc::cursor;
	if (pc::sidebar != NULL)
		delete pc::sidebar;
	if (pc::MissionsMapdata != NULL)
		delete pc::MissionsMapdata;
	if (p::dispatcher != NULL)
		delete p::dispatcher;
	if (p::aequeue != NULL)
		delete p::aequeue;
	if ( p::ccmap != NULL)
		delete p::ccmap;
	if (pc::Ai != NULL)
    	delete pc::Ai;
	if (pc::PauseMenu != NULL)
    	delete pc::PauseMenu;
	if (pc::imagepool != NULL)
    	delete pc::imagepool;
	if (pc::gfxeng != NULL)
		delete pc::gfxeng;
/*
	if (pc::sfxeng != NULL) 
		delete pc::sfxeng;
*/
	CleanConfig ();
	VFS_Destroy();

	pc::PauseMenu		= NULL;
	pc::Ai				= NULL;
	pc::input			= NULL;
	pc::cursor			= NULL;
	pc::sidebar 		= NULL;
	pc::MissionsMapdata	= NULL;
	p::dispatcher		= NULL;
	p::aequeue			= NULL;
	p::ccmap			= NULL;
	pc::imagepool		= NULL;
	pc::gfxeng			= NULL;
//	pc::sfxeng			= NULL;
}

void Game::InitializeGameClasses (void)
{
	// Initialise Video
	try {
		logger->note("Initialising the graphics engine...");
		pc::gfxeng = new GraphicsEngine();
		logger->note("done\n");
	} catch (GraphicsEngine::VideoError&) {
		logger->note("failed.  exiting\n");
		throw runtime_error("Unable to initialise the graphics engine");
	}
/*
	// Initialise Sound
	try {
		logger->note("Initialising the sound engine...");
		pc::sfxeng = new SoundEngine(pc::Config.nosound);
		logger->note("done\n");
	} catch (SoundEngine::SoundError&) {
		logger->note("failed.  exiting\n");
		throw runtime_error("Unable to initialise the sound engine");
	}
*/

	VFS_PreInit(pc::Config.binpath.c_str());
	VFS_Init(pc::Config.binpath.c_str());
	VFS_LoadGame(pc::Config.gamenum);

	// Get the mission maps
	pc::MissionsMapdata = new MissionMapsClass;
	
	// Init the image cache
	pc::imagepool = new std::vector<SHPImage*>();
	pc::imgcache->setImagePool(pc::imagepool);


	// Load the music files (for background music
	/// @TODO We've already loaded files.ini in the vfs.
	INIFile *fileini = GetConfig("files.ini");
	INIKey key = fileini->readIndexedKeyValue("general", pc::Config.gamenum, "PLAY");
	if (!pc::sfxeng->CreatePlaylist(key->second.c_str())) {
		logger->error("Could not create playlist!\n");
		throw GameError();
	}
}


void Game::HandleTiming (void)
{
    Uint32 uptime = p::aequeue->getElapsedTime();
    uptime /= 1000; // elapsed time in sec.
    if ( (uptime - OldUptime) > 6){
        HandleGlobalTrigger ( TRIGGER_EVENT_TIME_ELAPSED, uptime/6 );
        OldUptime = uptime;
    }
}

/** Play the mission. */
void Game::play()
{
bool				MissionWon = false;
MissionMapsClass	Missions;
string				MissionName;

	do{

		// Create the new game classes
		InitializeGameClasses ();

		// Draw the starting menu
		if (!MissionWon){
			// Play the battle control tirminated sound (if needed)
			if (BattleControlTerminated)
				pc::sfxeng->PlaySound("bct1.aud");
			HandleGameMenu();
		}

		BattleControlTerminated = false;

		// Exit the game (if the user wants to)
		if (pc::quit){
			printf ("Exit by user\n");
			FreeMemory();
			return;
		}

    	if (pc::Config.gamemode == GAME_MODE_SINGLE_PLAYER){
			if (pc::Config.mside == "gdi")
				Missions.GetGdiMissionMap (MissionNr, pc::Config.mapname);
			else if (pc::Config.mside == "nod")
				Missions.GetNodMissionMap (MissionNr, pc::Config.mapname);
			else{
				printf ("%s line %i: Unknown mission type\n", __FILE__, __LINE__);
			}
		}

		printf ("Mission name = %s\n", MissionName.c_str());

		// Initialize (load) the map
		InitializeMap( pc::Config.mapname );

		// Start playing the background music
		// pc::sfxeng->PlayTrack(p::ccmap->getMissionData().theme);

		// Setup the current game
		pc::gfxeng->setupCurrentGame();

		if (!pc::Config.UseFogOfWar){
			printf ("%s line %i: Use no fog of war?\n", __FILE__, __LINE__);
			p::ppool->getLPlayer()->setVisBuild(Player::SOB_SIGHT, true);
		}

    	if (p::ccmap->getGameMode() == GAME_MODE_SINGLE_PLAYER){
        	/// Jump to start location
	        /// @TODO: Jump to correct start location in multiplayer games.
    	    p::ccmap->restoreLocation(0);

	    }else{
			// The following is a hack that sets the location to the current players location
			Player *LocalPlayer = p::ppool->getLPlayer();

			Uint32 StartPos = LocalPlayer->getPlayerStart();

//			printf ("Startpos = %i, Current scrollpos = %i\n", StartPos, p::ccmap->getScrollPos());
			Uint32 x, y;

			if (StartPos%p::ccmap->getWidth() > 14)
				x = StartPos%p::ccmap->getWidth() - 14;
			else
				x = StartPos%p::ccmap->getWidth();

			if (StartPos/p::ccmap->getWidth() > 10)
				y = StartPos/p::ccmap->getWidth() - 10;
			else
				y = StartPos/p::ccmap->getWidth() - 10;


			if (x > p::ccmap->getWidth())
				x = p::ccmap->getWidth();

			if (y > p::ccmap->getHeight())
				y = p::ccmap->getHeight();

			p::ccmap->setScrollPos(x, y);
		}


//		printf ("Startpos = %i, Current scrollpos = %i\n", StartPos, p::ccmap->getScrollPos());

		OldUptime = 0;

	    /* main gameloop */
    	pc::Ai = new ai;
		pc::PauseMenu = new PauseMenuClass;
	    while (!pc::input->shouldQuit() && !pc::quit) {

			// Draw the pauze menu (when needed)
			while (pc::Config.pause)
				pc::PauseMenu->HandleMenu();

			// Draw the scene
			pc::gfxeng->renderScene();

			// Run scheduled events
			p::aequeue->runEvents();

			// Handle the input
			pc::input->handle();

			// Handle the ai
			pc::Ai->handle();

			// Handle timing triggers
			HandleTiming ();

			if (gamemode == 2) {
				// Synchronise events with server
			}
		}

		BattleControlTerminated = true;

		// Stop the music
		pc::sfxeng->StopMusic();

		if (gamemode == 0) {
			if (p::ppool->hasWon() ) {
				MissionWon = true;
				MissionNr++;
				try {
					VQAMovie *mov = new VQAMovie(p::ccmap->getMissionData().winmov);
					mov->play();
				} catch (std::runtime_error&) {
				}
			} else if (p::ppool->hasLost() ) {
				MissionWon = false;
				try {
					VQAMovie *mov = new VQAMovie(p::ccmap->getMissionData().losemov);
					mov->play();
				} catch (std::runtime_error&) {
				}
			}else{
				// Game was abborted
				MissionWon = false;
			}
		}

		// Stop all the music
		pc::sfxeng->StopMusic();
		pc::sfxeng->StopLoopedSound(-1);

		dumpstats();

		FreeMemory();

	}while (!pc::quit);
}

void Game::dumpstats()
{
    Player* pl;
    Uint8 h,m,s,i;
    Uint32 uptime = p::aequeue->getElapsedTime();
    uptime /= 1000;
    h = uptime/3600;
    uptime %= 3600;
    m = uptime/60;
    s = uptime%60;
    logger->renderGameMsg(false);
    logger->gameMsg("Time wasted: %i hour%s%i minute%s%i second%s",
                    h,(h!=1?"s ":" "),m,(m!=1?"s ":" "),s,(s!=1?"s ":" "));
    for (i = 0; i < p::ppool->getNumPlayers(); i++) {
        pl = p::ppool->getPlayer(i);
        logger->gameMsg("%s\nUnit kills:  %i\n     losses: %i\n"
                        "Structure kills:  %i\n          losses: %i\n",
                        pl->getName(),pl->getUnitKills(),pl->getUnitLosses(),
                        pl->getStructureKills(),pl->getStructureLosses());
    }
}
