/*****************************************************************************
 * sidebar.cpp - sidebar class
 *
 ****************************************************************************/
//#include <boost/mem_fn.hpp>
#include <cstdio>
#include <cstdlib>
#include <stdexcept>
#include <string>
#include "buildqueue.h"
#include "common.h"
#include "config.h"
#include "dispatcher.h"
#include "font.h"
#include "input.h"
#include "imagecache.h"
#include "logger.h"
#include "shpimage.h"
#include "sidebar.h"
#include "soundengine.h"
#include "playerpool.h"
#include "unitandstructurepool.h"
#include "unit.h"
#include "vfs.h"
#include "graphicsengine.h"
#include "cursor.h"

/// @TODO Move this into config file(s)
static const char* radarnames[] =
{
//   Gold Edition radars: NOD, GDI and Dinosaur
    "hradar.nod", "hradar.gdi", "hradar.jp",
//   Dos Edition radars
    "radar.nod",  "radar.gdi",  "radar.jp",
//   Red Alert radars (We only use large sidebar for RA)
/*    "ussrradr.shp", "natoradr.shp", "natoradr.shp"*/
    "uradrfrm.shp", "nradrfrm.shp", "nradrfrm.shp"
};

class ImageCache;

using std::string;
using std::runtime_error;
using std::vector;
using std::replace;


/**
 * @param pl local player object
 * @param height height of screen in pixels
 * @param theatre terrain type of the theatre
 *      TD: desert, temperate and winter
 *      RA: temperate and snow
 */
Sidebar::Sidebar(Player *pl, Uint16 height, const char *theatre)
    : radarlogo(0), tab(0), sbar(0), gamefnt(new Font(/*"grad6fnt.fnt"*/"scorefnt.fnt")), visible(true),
    vischanged(true), theatre(theatre), buttondown(0), bd(false),
    radaranimating(false), unitoff(0), structoff(0), player(pl), scaleq(-1)
{
    const char* tmpname;
    Uint8 side;
    SDL_Surface *tmp;

    // If we can't load these files, there's no point in proceeding, thus we let
    // the default handler for runtime_error in main() catch.
    if (pc::Config.gamenum == GAME_TD) {
        tmpname = VFS_getFirstExisting(2,"htabs.shp","tabs.shp");
        if (tmpname == NULL) {
            throw runtime_error("Unable to find the tab images! (Missing updatec.mix?)");
        } else if (strcasecmp(tmpname,"htabs.shp")==0) {
            isoriginaltype = false;
        } else {
            isoriginaltype = true;
        }
        try {
            tab = pc::imgcache->loadImage(tmpname, scaleq);
        } catch (ImageNotFound&) {
            throw runtime_error("Unable to load the tab images!");
        }
    } else {
        isoriginaltype = false;

	// Load tabs images
        try {
            tab = pc::imgcache->loadImage("tabs.shp", scaleq);
        } catch (ImageNotFound&) {
            throw runtime_error("Unable to load the tab images!");
        }

	// Load powerbar images
        try {
            powerbar = pc::imgcache->loadImage("powerbar.shp", scaleq);
        } catch (ImageNotFound&) {
            throw runtime_error("Unable to load the powerbar_up images!");
        }

	// Load power indicator image
        try {
            power_indicator = pc::imgcache->loadImage("power.shp", scaleq);
        } catch (ImageNotFound&) {
            throw runtime_error("Unable to load the power image!");
        }

	// Load the sell icon image
        try {
            sell_icon = pc::imgcache->loadImage("sell.shp", scaleq);
        } catch (ImageNotFound&) {
            throw runtime_error("Unable to load the sell icon!");
        }

	// Load the repair icon image
        try {
            repair_icon = pc::imgcache->loadImage("repair.shp", scaleq);
        } catch (ImageNotFound&) {
            throw runtime_error("Unable to load the repair icon!");
        }

	// Load the map icon image
        try {
            map_icon = pc::imgcache->loadImage("map.shp", scaleq);
        } catch (ImageNotFound&) {
            throw runtime_error("Unable to load the map icon!");
        }


	//Powerbar_up	= this->ReadShpImage ("powerbar.shp", 0);
	//Powerbar_down	= this->ReadShpImage ("powerbar.shp", 1);
	//PowerIndicator	= this->ReadShpImage ("power.shp", 0);
    }

    tmp = pc::imgcache->getImage(tab).image;

    tablocation.x = 0;
    tablocation.y = 0;
    tablocation.w = tmp->w;
    tablocation.h = tmp->h;

    spalnum = pl->getStructpalNum();
    if (!isoriginaltype) {
        spalnum = 0;
    }
    // Select a radar image based on sidebar type (DOS/Gold) and player side
    // (Good/Bad/Other)
    if ((player->getSide()&~PS_MULTI) == PS_BAD) {
        side = isoriginaltype?3:0;
    } else if ((player->getSide()&~PS_MULTI) == PS_GOOD) {
        side = isoriginaltype?4:1;
    } else {
        side = isoriginaltype?5:2;
    }

    if (pc::Config.gamenum == GAME_RA) {
        // RA follows (Gold names) * 3 and (DOS names) * 3
        side += 6;
    }

    radarname = radarnames[side];
    radaranim = 0;
    try {
        radaranimnumb = radarlogo = pc::imgcache->loadImage(radarname, scaleq);
    } catch (ImageNotFound&) {
        /// @TODO This problem should ripple up to the "game detection" layer
        // so it can try again from scratch with a different set of data files.
        logger->error("Hmm.. managed to misdetect sidebar type\n");
        try {
            // Switch between Gold and DOS
            if (side < 3) {
                side += 3;
            } else {
                side -= 3;
            }
            radarname = radarnames[side];
            radarlogo = pc::imgcache->loadImage(radarname, scaleq);
            isoriginaltype = !isoriginaltype;
        } catch (ImageNotFound&) {
            logger->error("Unable to load the radar-image! (Maybe you run c&c gold but have forgoten updatec.mix?)\n");
            throw SidebarError();
        }
    }
	if (pc::Config.gamenum == GAME_RA) {
		if ((player->getSide()&~PS_MULTI) == PS_BAD) {
			radaranimnumb = pc::imgcache->loadImage("ussrradr.shp", scaleq);
		}else
			radaranimnumb = pc::imgcache->loadImage("natoradr.shp", scaleq);
	}

    SDL_Surface *radar;
    if (pc::Config.gamenum == GAME_TD)
    	radar = pc::imgcache->getImage(radarlogo).image;
    else
    	radar = pc::imgcache->getImage(radarlogo,1).image;
//	radar = ReadShpImage ((char *)radarnames[side], 1);

    radarlocation.x = 0;  radarlocation.y = 0;
    radarlocation.w = radar->w;  radarlocation.h = radar->h;

    /// @TODO Move these values into config
    steps = ((pc::Config.gamenum == GAME_RA)?54:108);

    sbarlocation.x = sbarlocation.y = sbarlocation.w = sbarlocation.h = 0;

    std::fill(greyFixed, greyFixed+256, false);

    pc::sidebar = this;
    SetupButtons(height);

    repair_but_state = sell_but_state = map_but_state = 0;
}

Sidebar::~Sidebar()
{
    Uint32 i;

    SDL_FreeSurface(sbar);


    for (i=0; i < uniticons.size(); ++i) {
        delete[] uniticons[i];
    }

    for (i=0; i < structicons.size(); ++i) {
        delete[] structicons[i];
    }
}

SDL_Surface *Sidebar::ReadShpImage (char *Name, int ImageNumb)
{
SDL_Surface	*image,
		*shadow;
SHPImage	*TempPic;

	try {
		TempPic = new SHPImage(Name, -1);
	} catch (ImageNotFound&) {
		printf ("%s line %i: Image not found\n", __FILE__, __LINE__);
		return NULL;
	}

	printf ("%s line %i: %s has %i images\n", __FILE__, __LINE__, Name, TempPic->getNumImg());

	TempPic->getImage(ImageNumb, &image, &shadow, 1);

	delete TempPic;

	SDL_FreeSurface(shadow);

	return image;
}


/**
 * @returns whether the sidebar's visibility has changed.
 */
bool Sidebar::getVisChanged() {
    if (vischanged) {
        vischanged = false;
        return true;
    }
    return false;
}

void Sidebar::ToggleVisible()
{
    visible = !visible;
    vischanged = true;
}

void Sidebar::ReloadImages()
{
    // Free the sbar, so that it is redrawn on next update
    SDL_FreeSurface(sbar);
    sbar = NULL;
    sbarlocation.x = sbarlocation.y = sbarlocation.w = sbarlocation.h = 0;

    // Reload the font we use
    gamefnt->reload();

    // Reload all the buttons
//    for_each(buttons.begin(), buttons.end(), boost::mem_fn(&Sidebar::SidebarButton::ReloadImage));
	for (unsigned int i = 0; i < buttons.size(); i++){
		buttons[i]->ReloadImage();
	}

    // Invalidate grey fix table
    std::fill(greyFixed, greyFixed+256, false);
}

SDL_Surface *Sidebar::getSidebarImage(SDL_Rect location)
{
    SDL_Rect dest, src, newdest;
    SDL_Surface *temp;

    if (location.w == sbarlocation.w && location.h == sbarlocation.h)
        return sbar;

    SDL_FreeSurface(sbar);

    temp = SDL_CreateRGBSurface(SDL_SWSURFACE, location.w, location.h, 16, 0, 0, 0, 0);
    sbar = SDL_DisplayFormat(temp);
    SDL_FreeSurface(temp);
    location.x = 0;
    location.y = 0;

    /// @TODO HACK
    if (isoriginaltype || pc::Config.gamenum == GAME_RA) {
        SDL_FillRect(sbar, &location, SDL_MapRGB(sbar->format, 1, 1, 1));
    } else {
        try {
            // Get the index of the btexture
            Uint32 idx = pc::imgcache->loadImage("btexture.shp", scaleq);

            // Get the SDL_Surface for this texture
            SDL_Surface* texture = pc::imgcache->getImage(idx).image;

            dest.x = 0;
            dest.y = 0;
            dest.w = location.w;
            dest.h = texture->h;
            src.x = 0;
            src.y = 0;
            src.w = location.w;
            src.h = texture->h;
            for (dest.y = 0; dest.y < location.h; dest.y += dest.h) {
                SDL_BlitSurface(texture, &src, sbar, &dest);
            }
        } catch (ImageNotFound&) {
            logger->error("Unable to load the background texture\n");
            SDL_FillRect(sbar, &location, SDL_MapRGB(sbar->format, 0xa0, 0xa0, 0xa0));
        }
    }
    sbarlocation = location;

	SDL_Surface* radar;
	if (!Input::isMinimapEnabled()) {

		if (pc::Config.gamenum == GAME_TD)
			radar = pc::imgcache->getImage(radarlogo).image;
		else
			radar = pc::imgcache->getImage(radarlogo, 1).image;

		SDL_SetColorKey(radar,SDL_SRCCOLORKEY, 0xffffff);

		SDL_BlitSurface(radar, NULL, sbar, &radarlocation);

		// overdraw the with the correct upper part
		radar = pc::imgcache->getImage(radaranimnumb, 0).image;
		SDL_SetColorKey(radar,SDL_SRCCOLORKEY, 0xffffff);
		newdest.x = radarlocation.x;
		newdest.y = radarlocation.y;
		newdest.h = radar->h;
		newdest.w = radar->w;
//		SDL_FillRect(sbar, &newdest, SDL_MapRGB(pc::sidebar->sbar->format, 0x0a, 0x0a, 0x0a));
		SDL_BlitSurface( radar, NULL, sbar, &newdest);
	}else{

		if (pc::Config.gamenum == GAME_RA){
			radar = pc::imgcache->getImage(radarlogo, 1).image;
			SDL_SetColorKey(radar,SDL_SRCCOLORKEY, 0xffffff);
			SDL_BlitSurface(radar, NULL, sbar, &radarlocation);
		}
	}



    for (Sint8 x = buttons.size()-1;x>=0;--x) {
        DrawButton(x);
    }

    this->DrawPowerbar();
    this->DrawSpecialIcons();

    return sbar;
}

void Sidebar::AddButton(Uint16 x, Uint16 y, const char* fname, Uint8 f, Uint8 pal)
{
    SidebarButton *t(new SidebarButton(x, y, fname, f, theatre, pal));
    buttons.push_back(t);
    vischanged = true;
}

void Sidebar::SetupButtons(Uint16 height)
{
const char* tmpname;
Uint16 scrollbase;
Uint8 t;
int ButtonXpos;

	Uint32 startoffs;

	if (pc::Config.gamenum == GAME_TD){
		ButtonXpos	= 10;
		startoffs	= tablocation.h + radarlocation.h;
	}else{
		ButtonXpos	= 25;
		startoffs	= radarlocation.h + 5;
	}
	SHPImage *strip;

	tmpname = VFS_getFirstExisting(3,"stripna.shp","hstrip.shp","strip.shp");
	if (tmpname == 0) {
		logger->error("Unable to find strip images for sidebar, exiting\n");
		throw SidebarError();
	}
	try {
		strip = new SHPImage(tmpname, scaleq);
	} catch (ImageNotFound&) {
		logger->error("Unable to load strip images for sidebar, exiting\n");
		throw SidebarError();
	}

	geom.bh = strip->getHeight();
	geom.bw = strip->getWidth();

	delete strip;

	if (geom.bh > 100)
		geom.bh = geom.bh>>2;

	buildbut = ((height-startoffs)/geom.bh)-2;

	// Add the scroll up/down buttons
	if (pc::Config.gamenum == GAME_TD){
		startoffs += geom.bh;
		scrollbase = startoffs + geom.bh*buildbut;
		AddButton(ButtonXpos+geom.bw,scrollbase,"stripup.shp",sbo_scroll|sbo_unit|sbo_up,0); // 0
		AddButton(ButtonXpos,scrollbase,"stripup.shp",sbo_scroll|sbo_structure|sbo_up,0); // 1

		AddButton(ButtonXpos+geom.bw+(geom.bw>>1),scrollbase,"stripdn.shp",sbo_scroll|sbo_unit|sbo_down,0); // 2
		AddButton(ButtonXpos+(geom.bw>>1),scrollbase,"stripdn.shp",sbo_scroll|sbo_structure|sbo_down,0); // 3
	}else if (pc::Config.gamenum == GAME_RA){
		scrollbase = startoffs + geom.bh*buildbut;
		AddButton(ButtonXpos+geom.bw,scrollbase,"stripup.shp",sbo_scroll|sbo_unit|sbo_up,0); // 0
		AddButton(ButtonXpos,scrollbase,"stripup.shp",sbo_scroll|sbo_structure|sbo_up,0); // 1

		AddButton(ButtonXpos+geom.bw+(geom.bw>>1),scrollbase,"stripdn.shp",sbo_scroll|sbo_unit|sbo_down,0); // 2
		AddButton(ButtonXpos+(geom.bw>>1),scrollbase,"stripdn.shp",sbo_scroll|sbo_structure|sbo_down,0); // 3
	}

    // The order in which the AddButton calls are made MUST be preserved
    // Two loops are made so that all unit buttons and all structure buttons
    // are grouped contiguously (4,5,6,7,...) compared to (4,6,8,10,...)

    for (t=0;t<buildbut;++t) {
        if (pc::Config.gamenum == GAME_RA){
		if ((player->getSide()&~PS_MULTI) == PS_BAD)
			AddButton(ButtonXpos+geom.bw,startoffs+geom.bh*t,"stripus.shp",sbo_build|sbo_unit,0);
		else
			AddButton(ButtonXpos+geom.bw,startoffs+geom.bh*t,"stripna.shp",sbo_build|sbo_unit,0);
        }else
		AddButton(ButtonXpos+geom.bw,startoffs+geom.bh*t,"strip.shp",sbo_build|sbo_unit,0);
    }
    for (t=0;t<buildbut;++t) {
        if (pc::Config.gamenum == GAME_RA){
		if ((player->getSide()&~PS_MULTI) == PS_BAD)
			AddButton(ButtonXpos,startoffs+geom.bh*t,"stripus.shp",sbo_build|sbo_structure,spalnum);
		else
			AddButton(ButtonXpos,startoffs+geom.bh*t,"stripna.shp",sbo_build|sbo_structure,spalnum);
          }else
		AddButton(ButtonXpos,startoffs+geom.bh*t,"strip.shp",sbo_build|sbo_structure,spalnum);
    }

    if (pc::Config.gamenum == GAME_RA)
        startoffs += geom.bh;


    UpdateAvailableLists();
    UpdateIcons();
    if (uniticons.empty() && structicons.empty()) {
        visible = false;
        vischanged = true;
    }
}

Uint8 Sidebar::getButton(Uint16 x,Uint16 y)
{
    SDL_Rect tmp;

    for (Uint8 i=0;i<buttons.size();++i) {
        tmp = buttons[i]->getRect();
        if (x>=tmp.x && y>=tmp.y && x<(tmp.x+tmp.w) && y<(tmp.y+tmp.h)) {
            return i;
        }
    }
    return 255;
}

void Sidebar::GetButtonName(Uint8 index, char *UnitOrStructureName)
{
Uint8 function = buttons[index]->getFunction();
Uint8* offptr;
std::vector<char*>* vecptr;
Uint8 type;

	strncpy(UnitOrStructureName,"xxxx",4);

	UnitOrStructureName[5] = '\0';

	// Check that this is a build button and not a scroll button
	if ((function&0x3) != 1)
		return;

	Uint8 VecPtrIndex = index - 3 - ((function&sbo_unit)?0:buildbut);

	type = (function&sbo_unit);

	if (type) {
		offptr = &unitoff;
		vecptr = &uniticons;
	} else {
		offptr = &structoff;
		vecptr = &structicons;
	}

	if ( (unsigned)(*vecptr).size() > ((unsigned)(*offptr)+VecPtrIndex - 1) ) {
		strncpy(UnitOrStructureName,(*vecptr)[(*offptr+VecPtrIndex-1)],13);
		UnitOrStructureName[strlen((*vecptr)[(*offptr+VecPtrIndex-1)])-8] = 0x0;
	} else {
		// Out of range
		return;
	}
}

void Sidebar::DrawButtonTooltip(Uint8 index)
{
char 			UnitOrStructureName[20];	// should be at least 13 :)
Uint8 			unit;
Uint8 			function = buttons[index]->getFunction();
//char			TooltipText[20];
std::ostringstream	TipString;

	GetButtonName(index, UnitOrStructureName);

	unit = ( function & sbo_unit );

	UnitOrStructureType* Type;
	if (unit) {
		Type = p::uspool->getUnitTypeByName(UnitOrStructureName);
	} else {
		Type = p::uspool->getStructureTypeByName(UnitOrStructureName);
	}

	if (Type == NULL)
		return;

	TipString.str("");
	TipString<<"$"<<(unsigned int)(Type->getCost());

	pc::cursor->setTooltip (TipString.str());
}

void Sidebar::DrawButton(Uint8 index)
{
SDL_Rect src;
//	printf ("%s line %i: Draw buttons\n", __FILE__, __LINE__);

	if (sbar == 0) {
		/// @TODO Ensure we don't actually get called when this is the case
		return;
	}
	SDL_Rect dest = buttons[index]->getRect();
//	dest.y += 10;
	SDL_FillRect(sbar, &dest, 0x0);
	Uint8 func = buttons[index]->getFunction();

	// Blit the button's image onto the sidebar surface
	SDL_Surface *temp = buttons[index]->getSurface();
 	if (temp->h > 100) {
		int ButtonYpos = (dest.y - radarlocation.h - 5)/*/dest.h*/;
		//printf ("ButtonYpos = %i\n", ButtonYpos);

		src.x = 0;
		src.y = ButtonYpos/* * dest.h*/;
		src.w = temp->w;
		src.h = dest.h;
		SDL_BlitSurface(temp, &src, sbar, &dest);
	}else
    		SDL_BlitSurface(temp, NULL, sbar, &dest);

	vischanged = true;
	// Skip scroll buttons
	if (index < 4) {
		return;
	}
	// Calculate which icon was clicked on
	Uint8 offset = index - 4 + // First four buttons are scroll buttons
			((func&sbo_unit)
			? unitoff  // Unit buttons are created first
			: (structoff-buildbut) // See comment around line 265
			);
	vector<char*>& icons = ((func&sbo_unit)
				?(uniticons)
				:(structicons));

	if (offset >= icons.size()) {
		return;
	}

    // Extract type name from icon name, e.g. NUKE from NUKEICON.SHP
    Uint32 length = strlen(icons[offset])-8; if (length>13) length = 13;
    string name(icons[offset], length);

    static const char* stat_mesg[] = {
        "???",		// BQ_INVALID
        "",			// BQ_EMPTY
        "",			// BQ_RUNNING
        "hold",		// BQ_PAUSED
        "ready",	// BQ_READY
		"",// BQ_CANCELLED
		""			// BQ_ALL_PAUSED
    };
    static struct {Uint32 x,y;} stat_pos[5];
    static bool posinit = false;
    if (!posinit) {
        for (Uint8 i = 0; i < 5; ++i) {
            // Here we assume that all sidebar icons are the same size
            // (which is pretty safe)
            stat_pos[i].x = (dest.w - getFont()->calcTextWidth(stat_mesg[i])) >> 1;
            stat_pos[i].y = dest.h - 15;
        }
        posinit = true;
    }
    ConStatus status;
    Uint8 quantity, progress, imgnum;

    UnitOrStructureType* type;
    if (func&sbo_unit) {
        type = p::uspool->getUnitTypeByName(name.c_str());
    } else {
        type = (UnitOrStructureType*)p::uspool->getStructureTypeByName(name.c_str());
    }
    if (0 == type) {
        throw runtime_error("Asking for \""+name+"\" resulted in a null pointer");
    }
    status = player->getStatus(type, &quantity, &progress);
    if (BQ_INVALID == status) {
        getFont()->drawText(stat_mesg[status], sbar, dest.x + stat_pos[status].x, dest.y + stat_pos[status].y);
        // Grey out invalid items for prettyness
        DrawClock(index, 0);
        return;
    }
    imgnum = (steps*progress)/100;
    if (100 != progress) {
        DrawClock(index, imgnum);
    }
    if (quantity > 0) {
        char tmp[4];
        sprintf(tmp, "%i", quantity);
        getFont()->drawText(tmp, sbar, dest.x+3, dest.y+3);
    }
    /// @TODO This doesn't work when you immediately pause after starting to
    // build (but never draws the clock for things not being built).
    if (progress > 0) {
        getFont()->drawText(stat_mesg[status], sbar, dest.x + stat_pos[status].x, dest.y + stat_pos[status].y);
    }

}

void Sidebar::DrawPowerbar(void)
{
SDL_Rect dest;
SDL_Surface	*Powerbar_up,
		*Powerbar_down
		/*,*PowerIndicator*/;

	if (pc::Config.gamenum != GAME_RA || sbar == 0) 
		return;

	Powerbar_up = pc::imgcache->getImage(powerbar,0).image;
	dest.x	= 0;
	dest.y	= radarlocation.h;
	dest.w	= Powerbar_up->w;
	dest.h	= Powerbar_up->h;
	SDL_BlitSurface(Powerbar_up, NULL, sbar, &dest);

	Powerbar_down = pc::imgcache->getImage(powerbar,1).image;
	dest.x	= 0;
	dest.y	= radarlocation.h + Powerbar_up->h;
	dest.w	= Powerbar_down->w;
	dest.h	= Powerbar_down->h;	
	SDL_BlitSurface(Powerbar_down, NULL, sbar, &dest);

}

void Sidebar::UpdatePowerbar(void)
{
static unsigned int	MaxPower = 500;
SDL_Rect		dest;
SDL_Surface		*Powerbar_up,
			*Powerbar_down,
			*PowerIndicator;
Uint32			IndicatorBarColor, IndicatorBarColorDark;
static unsigned int 	OldPower, OldPowerused;
static bool		LowPowerSoundPlayed = false;

	if (sbar == 0 || pc::Config.gamenum != GAME_RA)
		return;

	// Get the power status
	Player *lplayer = p::ppool->getLPlayer();
	unsigned int power = lplayer->getPower();
	unsigned int powerused = lplayer->getPowerUsed();

	// Check for change
	if (OldPower == power && OldPowerused	== powerused)
		return;

	// Handle playing the low power sound
	if (powerused > power-(power * 0.2) &&  powerused < power){
		if (!LowPowerSoundPlayed)
			pc::sfxeng->PlaySound("lopower1.aud");
		LowPowerSoundPlayed = true;
	}else{
		LowPowerSoundPlayed = false;
	}



//	printf ("Update powerbar\n");

	Powerbar_up = pc::imgcache->getImage(powerbar,0).image;
	Powerbar_down = pc::imgcache->getImage(powerbar,1).image;
	PowerIndicator = pc::imgcache->getImage(power_indicator,0).image;



	// Redraw the powerbar upper part (so it is clean again)
	dest.x	= 0;
	dest.y	= radarlocation.h;
	dest.w	= Powerbar_up->w;
	dest.h	= Powerbar_up->h;
	SDL_BlitSurface(Powerbar_up, NULL, sbar, &dest);

	// Redraw the powerbar lower part (so it is clean again)
	dest.x	= 0;
	dest.y	= radarlocation.h + Powerbar_up->h;
	dest.w	= Powerbar_down->w;
	dest.h	= Powerbar_down->h;	
	SDL_BlitSurface(Powerbar_down, NULL, sbar, &dest);

	if (power > MaxPower)
		MaxPower = power * 4 + 500;

	if (powerused > MaxPower)
		MaxPower = powerused * 4 + 500;


	unsigned char PowerPercentage	= power * 100/ MaxPower;
	unsigned int MaxPowerHeight	= Powerbar_up->h + Powerbar_down->h - 31 - 46;
	unsigned int PowerHeight	= MaxPowerHeight * PowerPercentage/100;

	dest.x	= 11;
	dest.y	= radarlocation.h + 31 + MaxPowerHeight - PowerHeight;
	dest.w	= 2;
	dest.h	= PowerHeight;

	//
	// Handle the power indicator bar color
	//
	if (powerused > power-(power * 0.2) &&  powerused < power){
		// low power
		IndicatorBarColor	= SDL_MapRGB(sbar->format, 0xff, 0xa0, 0);
		IndicatorBarColorDark	= SDL_MapRGB(sbar->format, 0xff, 0x17, 0);
	}else if (powerused > power){
		// power off
		IndicatorBarColor	= SDL_MapRGB(sbar->format, 0xff, 0, 0);
		IndicatorBarColorDark	= SDL_MapRGB(sbar->format, 0xaa, 0, 0);
	}else{
		// power oke
		IndicatorBarColor	= SDL_MapRGB(sbar->format, 0, 0xff, 0);
		IndicatorBarColorDark	= SDL_MapRGB(sbar->format, 0, 0xaa, 0);
	}

	SDL_FillRect(sbar, &dest, IndicatorBarColor);
	dest.x	+= 2;
	SDL_FillRect(sbar, &dest, IndicatorBarColorDark);

	OldPower	= power;
	OldPowerused	= powerused;

	// Draw the power used indicator
	
	dest.x	= 2;
	dest.y	= radarlocation.h + 31 + MaxPowerHeight - (MaxPowerHeight * powerused / MaxPower) - PowerIndicator->h/2;
	dest.w	= PowerIndicator->w;
	dest.h	= PowerIndicator->h;
	SDL_BlitSurface(PowerIndicator, NULL, sbar, &dest);
//
}

#define S_BUTTON_NONE	255
#define S_BUTTON_REPAIR	1
#define S_BUTTON_SELL	2
#define S_BUTTON_WORLD	3
Uint8 Sidebar::getSpecialButton(Uint16 x,Uint16 y)
{

//RepairLoc, SellLoc, MapLoc;


        if (x>=RepairLoc.x && y>=RepairLoc.y && x<(RepairLoc.x+RepairLoc.w) && y<(RepairLoc.y+RepairLoc.h)) {
            return S_BUTTON_REPAIR;
        }else if (x>=SellLoc.x && y>=SellLoc.y && x<(SellLoc.x+SellLoc.w) && y<(SellLoc.y+SellLoc.h)) {
            return S_BUTTON_SELL;
	}else if (x>=MapLoc.x && y>=MapLoc.y && x<(MapLoc.x+MapLoc.w) && y<(MapLoc.y+MapLoc.h)) {
            return S_BUTTON_WORLD;
	}

    return S_BUTTON_NONE;
}

void Sidebar::setSpecialButtonState(Uint8 Button, Uint8 State)
{

//RepairLoc, SellLoc, MapLoc;
	switch (Button){
		case S_BUTTON_REPAIR:
			if (State < 3)
				repair_but_state = State;
			break;
		case S_BUTTON_SELL:
			if (State < 3)
				sell_but_state = State;
			break;
		case S_BUTTON_WORLD:
			if (State < 3)
				map_but_state = State;
			break;

	}
	DrawSpecialIcons();
}

Uint8 Sidebar::getSpecialButtonState(Uint8 Button)
{

//RepairLoc, SellLoc, MapLoc;
	switch (Button){
		case S_BUTTON_REPAIR:
			return repair_but_state;
			break;
		case S_BUTTON_SELL:
			return sell_but_state;
			break;
		case S_BUTTON_WORLD:
			return map_but_state;
			break;

	}
	return false;
}



void Sidebar::DrawSpecialIcons(void)
{
//SDL_Rect	dest;
SDL_Surface	*Repair,
		*Sell,
		*Map;
int		IconSpacing = 45;

	if (sbar == 0) {
		/// @TODO Ensure we don't actually get called when this is the case
		return;
	}

	radarlocation.h = 160;

	if (pc::Config.gamenum == GAME_RA){

		if (repair_but_state < 3)
			Repair	= pc::imgcache->getImage(repair_icon,repair_but_state).image;
		else
			Repair	= pc::imgcache->getImage(repair_icon,0).image;
		RepairLoc.x	= 18;
		RepairLoc.y	= radarlocation.h - Repair->h +2;
		RepairLoc.w	= Repair->w;
		RepairLoc.h	= Repair->h;	
		SDL_BlitSurface(Repair, NULL, sbar, &RepairLoc);

		if (sell_but_state < 3)
			Sell	= pc::imgcache->getImage(sell_icon,sell_but_state).image;
		else
			Sell	= pc::imgcache->getImage(sell_icon,0).image;

		SellLoc.x	= 18+IconSpacing;
		SellLoc.y	= radarlocation.h - Sell->h +2;
		SellLoc.w	= Sell->w;
		SellLoc.h	= Sell->h;	
		SDL_BlitSurface(Sell, NULL, sbar, &SellLoc);

		if (map_but_state < 3)
			Map	= pc::imgcache->getImage(map_icon,map_but_state).image;
		else
			Map	= pc::imgcache->getImage(map_icon,3).image;

		MapLoc.x	= 18+2*IconSpacing;
		MapLoc.y	= radarlocation.h - Map->h +2;
		MapLoc.w	= Map->w;
		MapLoc.h	= Map->h;
		SDL_BlitSurface(Map, NULL, sbar, &MapLoc);
	}
}

void Sidebar::FixGrey(SDL_Surface* gr, Uint8 num)
{
    if (greyFixed[num]) {
        return;
    }

    SDL_LockSurface(gr);
    if(gr->format->BytesPerPixel == 4) {
        Uint32 *pixels = (Uint32 *)gr->pixels;
        Uint32 size = ((gr->h-1) * (gr->pitch >> 2)) + gr->w-1;
        replace(pixels, pixels+size, 0xa800U, 0x10101U);
    } else if(gr->format->BytesPerPixel == 2) {
        Uint16 *pixels = (Uint16 *)gr->pixels;
        Uint32 size = ((gr->h-1) * (gr->pitch >> 1)) + gr->w-1;
        replace(pixels, pixels+size, 0x540U, 0x21U);
    }
    SDL_UnlockSurface(gr);
    greyFixed[num] = true;
}

void Sidebar::DrawClock(Uint8 index, Uint8 imgnum) {
    Uint32 num = 0;
    try {
        // @TODO Move this check to config object
        if (pc::Config.gamenum == GAME_RA || isoriginaltype) {
            num = pc::imgcache->loadImage("clock.shp");
        } else {
            num = pc::imgcache->loadImage("hclock.shp");
        }
    } catch (ImageNotFound& e) {
        logger->error("Unable to load clock image!\n");
        return;
    }
    num += imgnum;
    num |= spalnum<<11;

    ImageCacheEntry& ice = pc::imgcache->getImage(num);
    SDL_Surface* gr = ice.image;
    FixGrey(gr, imgnum);
    SDL_SetAlpha(gr, SDL_SRCALPHA, 128);

    SDL_Rect dest = buttons[index]->getRect();

    SDL_BlitSurface(gr, NULL, sbar, &dest);
}

/** Event handler for clicking buttons.
 *
 * @bug This is an evil hack that should be replaced by something more flexible.
 */
/// @TODO Bleh, this function needs loving.
void Sidebar::ClickButton(Uint8 index, char* unitname, createmode_t* createmode) {
    Uint8 f = buttons[index]->getFunction();
    *createmode = CM_INVALID;
    switch (f&0x3) {
    case 0:
        return;
        break;
    case 1: // build
        Build(index - 3 - ((f&sbo_unit)?0:buildbut),(f&sbo_unit),unitname, createmode);
        break;
    case 2: // scroll
        DownButton(index);
        ScrollBuildList((f&sbo_up), (f&sbo_unit));
        break;
    default:
        logger->error("Sidebar::ClickButton. This should not happen (%i)\n",f&0x3);
        break;
    }
}

/// @TODO Bleh, this function needs loving.
void Sidebar::Build(Uint8 index, Uint8 type, char* unitname, createmode_t* createmode) {
    Uint8* offptr;
    std::vector<char*>* vecptr;

    if (type) {
        offptr = &unitoff;
        vecptr = &uniticons;
    } else {
        offptr = &structoff;
        vecptr = &structicons;
        *createmode = CM_STRUCT;
    }

    if ( (unsigned)(*vecptr).size() > ((unsigned)(*offptr)+index - 1) ) {
        strncpy(unitname,(*vecptr)[(*offptr+index-1)],13);
        unitname[strlen((*vecptr)[(*offptr+index-1)])-8] = 0x0;
    } else {
        // Out of range
        return;
    }

    if (*createmode != CM_STRUCT) {
        // createmode was set to CM_INVALID in the caller of this function
        UnitType* utype = p::uspool->getUnitTypeByName(unitname);
        if (0 != utype) {
            *createmode = (createmode_t)utype->getType();
        }
    }
}

void Sidebar::ScrollBuildList(Uint8 dir, Uint8 type) {
    Uint8* offptr;
    std::vector<char*>* vecptr;

    if (type) {
        offptr = &unitoff;
        vecptr = &uniticons;
    } else {
        offptr = &structoff;
        vecptr = &structicons;
    }

    if (dir) { //up
        if (*offptr>0) {
            --(*offptr);
        }
    } else { // down
        if ((unsigned)(*offptr+buildbut) < (unsigned)vecptr->size()) {
            ++(*offptr);
        }
    }

    UpdateIcons();
}

void Sidebar::UpdateSidebar() 
{
//	printf ("%s line %i: Update sidebar\n", __FILE__, __LINE__);	
	UpdateAvailableLists();
	UpdateIcons();
	UpdatePowerbar();
}

/** Rebuild the list of icons that are available.
 *
 * @bug Newer items should be appended, although with some grouping (i.e. keep
 * infantry at top, vehicles, aircraft, boats, then superweapons).
 */
void Sidebar::UpdateAvailableLists() {
    std::vector<const char*> units_avail, structs_avail;
    char* nametemp;
    bool played_sound;
    Uint8 prev_units_avail, prev_structs_avail;
    Uint32 i;
    played_sound = false;
    units_avail = p::uspool->getBuildableUnits(player);
    structs_avail = p::uspool->getBuildableStructures(player);
    prev_units_avail = uniticons.size();
    prev_structs_avail = structicons.size();
    if (units_avail.size() != uniticons.size()) {
        unitoff = 0;
        if (units_avail.size() > uniticons.size()) {
            played_sound = true;
            pc::sfxeng->PlaySound("newopt1.aud");
        }
        for (i=0;i<uniticons.size();++i) {
            delete[] uniticons[i];
        }
        uniticons.resize(0);
        for (i=0;i<units_avail.size();++i) {
            nametemp = new char[13];
            memset(nametemp,0x0,13);
            sprintf(nametemp,"%sICON.SHP",units_avail[i]);
            uniticons.push_back(nametemp);
        }
    }
    if (structs_avail.size() != structicons.size()) {
        structoff = 0;
        if ((structs_avail.size() > structicons.size())&&!played_sound) {
            pc::sfxeng->PlaySound("newopt1.aud");
        }
        for (i=0;i<structicons.size();++i) {
            delete[] structicons[i];
        }
        structicons.resize(0);
        for (i=0;i<structs_avail.size();++i) {
            nametemp = new char[13];
            memset(nametemp,0x0,13);
            sprintf(nametemp,"%sICON.SHP",structs_avail[i]);
            structicons.push_back(nametemp);
        }
    }
    visible = (visible || uniticons.size() > prev_units_avail ||
            structicons.size() > prev_structs_avail);
    if (uniticons.empty() && structicons.empty()) {
        visible = false;
    }
    vischanged = true;
}

/** Sets the images of the visible icons, having scrolled.
 * @TODO Provide a way to only update certain icons
 */
void Sidebar::UpdateIcons() {
    Uint8 i;

	// Unit buttons
	for (i=0;i<buildbut;++i) {
		if ((unsigned)(i+unitoff)>=(unsigned)uniticons.size()) {
			if (pc::Config.gamenum == GAME_RA){
				if ((player->getSide()&~PS_MULTI) == PS_BAD)
					buttons[4+i]->ChangeImage("stripus.shp", 0, RIGHT_SIDE_BUTTON);
				else
					buttons[4+i]->ChangeImage("stripna.shp", 0, RIGHT_SIDE_BUTTON);
			}else 
				buttons[4+i]->ChangeImage("strip.shp");
		} else {
			buttons[4+i]->ChangeImage(uniticons[i+unitoff]);
		}
	}

	// Structure buttons
	for (i=0;i<buildbut;++i) {
		if ((unsigned)(i+structoff)>=(unsigned)structicons.size()) {
			if (pc::Config.gamenum == GAME_RA){
				if ((player->getSide()&~PS_MULTI) == PS_BAD)
					buttons[buildbut+4+i]->ChangeImage("stripus.shp", 0, LEFT_SIDE_BUTTON);
				else
					buttons[buildbut+4+i]->ChangeImage("stripna.shp", 0, LEFT_SIDE_BUTTON);
			}else 
				buttons[buildbut+4+i]->ChangeImage("strip.shp");
		} else {
			buttons[buildbut+4+i]->ChangeImage(structicons[i+structoff]);
		}
	}

	for (Sint8 x=buttons.size()-1;x>=0;--x) {
		DrawButton(x);
	}
}

void Sidebar::DownButton(Uint8 index)
{
    if (index>3)
        return; // not a scroll button

    buttondown = index;

    bd = true;
    if (index == 0 || index == 1)
        buttons[index]->ChangeImage("stripup.shp",1);
    else
        buttons[index]->ChangeImage("stripdn.shp",1);
    UpdateIcons();

}

void Sidebar::ResetButton()
{
    if (!bd)
        return;
    bd = false;

    if (buttondown < 2)
        buttons[buttondown]->ChangeImage("stripup.shp",0);
    else
        buttons[buttondown]->ChangeImage("stripdn.shp",0);
    buttondown = 0;
    UpdateIcons();
}

void Sidebar::StartRadarAnim(Uint8 mode, bool* minienable)
{
    if (radaranimating == false && radaranim == NULL && sbar != NULL) {
        radaranimating = true;
        radaranim = new RadarAnimEvent(mode, minienable, radaranimnumb);
    }
}

void Sidebar::ScrollSidebar(bool scrollup)
{
    ScrollBuildList(scrollup, 0);
    ScrollBuildList(scrollup, 1);
}

Sidebar::SidebarButton::SidebarButton(Sint16 x, Sint16 y, const char* picname,
        Uint8 f, const char* theatre, Uint8 pal)
 : pic(0), function(f), palnum(pal), theatre(theatre), using_fallback(false)
{
    picloc.x = x;
    picloc.y = y;
    ChangeImage(picname);
}

void Sidebar::SidebarButton::ChangeImage(const char* fname)
{
    ChangeImage(fname, 0, LEFT_SIDE_BUTTON);
}

void Sidebar::SidebarButton::ChangeImage(const char* fname, Uint8 number)
{
    ChangeImage(fname, number, LEFT_SIDE_BUTTON);
}

SDL_Surface* Sidebar::SidebarButton::Fallback(const char* fname)
{
    SDL_Surface* ret;
    Uint32 width, height;
    Uint32 slen = strlen(fname);
    char* iname = new char[slen-7];

    using_fallback = true; // Ensures that the surface created gets destroyed later.
    fallbackfname = fname;

    strncpy(iname,fname,slen-8);
    iname[slen-8] = 0;
    width = pc::sidebar->getGeom().bw;
    height = pc::sidebar->getGeom().bh;
    ret = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_SRCCOLORKEY, width, height, 16, 0, 0, 0, 0);
    SDL_FillRect(ret, NULL, 0);
    pc::sidebar->getFont()->drawText(iname,ret,0,0);
    delete[] iname;
    return ret;
}

void Sidebar::SidebarButton::ChangeImage(const char* fname, Uint8 number, Uint8 side)
{
const char* name;
char goldname[32];
Uint32 slen = strlen(fname);

	// Error checking
	if (side != LEFT_SIDE_BUTTON && side != RIGHT_SIDE_BUTTON)
		return;

	if (using_fallback) {
		SDL_FreeSurface(pic);
		using_fallback = false;
	}

	if (pc::sidebar->isOriginalType() || pc::Config.gamenum == GAME_RA) {
		name = fname;
	} else {

		if (slen>8 && strcasecmp("icon.shp", fname+slen-8)==0) {
			strcpy(goldname, fname);
			goldname[slen-6] = 'N';
			goldname[slen-5] = 'H';
			goldname[slen-3] = theatre[0];
			goldname[slen-2] = theatre[1];
			goldname[slen-1] = theatre[2];
		} else {
			sprintf(goldname, "H%s", fname);
		}
		name = goldname;
	}

	try {
		picnum = pc::imgcache->loadImage(name);
		picnum += number;
		picnum |= palnum<<11;
		if (pc::Config.gamenum == GAME_TD){
			pic = pc::imgcache->getImage(picnum).image;
		}else if (pc::Config.gamenum == GAME_RA){
			if (side == LEFT_SIDE_BUTTON)
				pic = pc::imgcache->getImage(picnum, 0).image;
			else 
				pic = pc::imgcache->getImage(picnum, 1).image;
		}

	} catch(ImageNotFound&) {
		/* This is only a temporary fix; the real solution is to add a creation
		* function to the imagecache, which would ahandle creation and would mean
		* the imagecache would be responsible for deletion.  A nice side effect is
		* that those changes will be acompanied with a way to index created images
		* so we only create the fallback images once. */
		pic = Fallback(fname);
	}

	picloc.w = pic->w;
	picloc.h = pic->h;

	if (picloc.h > 100) {
		picloc.h >>=2;
		//pic->h >>= 2;
	}
}

void Sidebar::SidebarButton::ReloadImage()
{
    //If we are using fallback we must redraw
    if (using_fallback) {
        SDL_FreeSurface(pic);
        pic = Fallback(fallbackfname);
    } else {
        //else the image cache takes care of it
        pic = pc::imgcache->getImage(picnum).image;
    }
}

Sidebar::SidebarButton::~SidebarButton()
{
    if (using_fallback) {
        SDL_FreeSurface(pic);
        using_fallback = false;
    }
}


Sidebar::RadarAnimEvent::RadarAnimEvent(Uint8 mode, bool* minienable, Uint32 radar)
    : ActionEvent(1), mode(mode), minienable(minienable), radar(radar)
{
    switch (mode) {
    case 0:
        frame = 0;
        framend = 20;
        break;
    case 1:
        //*minienable = false;
        frame = 20;
        framend = 30;
        break;
    default:
        frame = 0;
    }


	if (pc::sidebar->sbar != NULL) {
		// Draw grey box where radar was
		// SDL_FillRect(pc::sidebar->sbar, dest, SDL_MapRGB(pc::sidebar->sbar->format, 0x0a, 0x0a, 0x0a));

		if (pc::Config.gamenum == GAME_TD)
			sdlRadar = pc::imgcache->getImage(pc::sidebar->radarlogo).image;
		else
			sdlRadar = pc::imgcache->getImage(pc::sidebar->radarlogo,1).image;
            }
    p::aequeue->scheduleEvent(this);
}

void Sidebar::RadarAnimEvent::run()
{
	//SDL_Rect *dest = &pc::sidebar->radarlocation;
	SDL_Surface *radarFrame;
	SDL_Rect dest;
	dest.x = pc::sidebar->radarlocation.x;
	dest.y = pc::sidebar->radarlocation.y;
	dest.h = pc::sidebar->radarlocation.h;
	dest.w = pc::sidebar->radarlocation.w;

    if (frame <= framend) {

        //If the sidebar is null don't even bother
        if (pc::sidebar->sbar != NULL) {
            // Draw radar loading frame
		radarFrame = pc::imgcache->getImage(radar, frame).image;
		dest.h = radarFrame->h;
		//SDL_FillRect(pc::sidebar->sbar, &dest, SDL_MapRGB(pc::sidebar->sbar->format, 0x0a, 0x0a, 0x0a));
		SDL_BlitSurface( radarFrame, NULL, pc::sidebar->sbar, &dest);
        }

        ++frame;
        p::aequeue->scheduleEvent(this);
    	pc::sidebar->DrawSpecialIcons();
    } else {
        if (mode == 0) {
		// We got radar
		if (pc::sidebar->sbar != NULL) {
			printf ("%s line %i: Draw last radar image (radar up, mode = %i)\n", __FILE__, __LINE__, mode);
			radarFrame = pc::imgcache->getImage(pc::sidebar->radarlogo, 1).image;
			SDL_SetColorKey(radarFrame,SDL_SRCCOLORKEY,  0xffffff);
			dest.h = radarFrame->h;
			SDL_BlitSurface( radarFrame, NULL, pc::sidebar->sbar, &dest);
		}
		*minienable = true;
	}else{
		// We don't have radar anymore
		if (pc::sidebar->sbar != NULL) {
			printf ("%s line %i: Draw last radar image (radar down, mode = %i)\n", __FILE__, __LINE__, mode);
			radarFrame = pc::imgcache->getImage(radar, 0).image;
			SDL_SetColorKey(radarFrame,SDL_SRCCOLORKEY, 0xffffff);
			dest.h = radarFrame->h;
			SDL_BlitSurface( radarFrame, NULL, pc::sidebar->sbar, &dest);
		}
		*minienable = false;
	}
        pc::sidebar->radaranimating = false;
	pc::sidebar->radaranim = NULL;
    	pc::sidebar->DrawSpecialIcons();
        delete this;
        return;
    }
}
