#ifndef WIN32
#include <arpa/inet.h>
#endif
#include <cassert>
#include <cctype>
#include <stdexcept>
#include <string>
#include "common.h"
#include "compression.h"
#include "fcnc_endian.h"
#include "graphicsengine.h"
#include "imageproc.h"
#include "inifile.h"
#include "logger.h"
#include "soundcommon.h"
#include "soundengine.h"
#include "soundfile.h"
#include "vfs.h"
#include "vqa.h"

//#define DEBUG_VQA

using std::runtime_error; using std::string;

namespace VQAPriv
{
// Copied from XCC Mixer (xcc.ra2.mods) by Olaf van der Spek

/// @TODO Fix/verify this for Big Endian

inline Uint8 VQA_HI_BYTE(Uint32 x) {
    return (x & 0xff000000) >> 24;
}

const Uint32 vqa_t_mask = 0x00ffffff;

// Each of these constants is derived from e.g. *(Uint32*)"CBF\0" etc.
// We can't have these in the above form if we want to use them as switch
// labels, even though they're consts.....
const Uint32 vqa_cbf_id = 0x464243;
const Uint32 vqa_cbp_id = 0x504243;
const Uint32 vqa_cpl_id = 0x4C5043;
const Uint32 vqa_snd_id = 0x444E53;
const Uint32 vqa_vpt_id = 0x545056;
const Uint32 vqa_vpr_id = 0x525056;
const Uint32 vqa_vqfl_id = 0x4C465156;
const Uint32 vqa_vqfr_id = 0x52465156;
}

using namespace VQAPriv;

/// @TODO We really need a glossary for all the VQA jargon
// VQA: Vector Quantised Animation
// SND: Sound
// FINF: Offsets (for what?) Info
// FORM: Contains header data
// CBP:
// CBF:
// CPL: Colour Palette (?)
// VPT: ??? - Last chunk in a frame
// VQFR: VQ Frame (?)

/**
 * @param the name of the vqamovie.
 */
VQAMovie::VQAMovie(const char* filename) : vqafile(0), CBF_LookUp(0),
        CBP_LookUp(0), VPT_Table(0), offsets(0), sndbuf(0)
{
    string fname(filename);

    if (toupper(filename[0]) != 'X') {
        fname += ".VQA";
        vqafile = VFS_Open(fname.c_str());
    }

    if (vqafile == 0) {
        throw runtime_error("No VQA file");
    }
    // Get header information for the vqa.  If the header is corrupt, we can die now.
    vqafile->seekSet(0);
    if (!DecodeFORMChunk()) {
        logger->error("VQA: Invalid FORM chunk\n");
        throw runtime_error("VQA: Invalid FORM chunk\n");
    }

#ifdef DEBUG_VQA
	printf ("%s line %i: File name\t\t\t\t= %s\n", __FILE__, __LINE__, filename);
	printf ("%s line %i: Header signature\t\t\t\t= %c%c%c%c%c%c%c%c\n", __FILE__, __LINE__, header.Signature[0], header.Signature[1], header.Signature[2], header.Signature[3], header.Signature[4], header.Signature[5], header.Signature[6], header.Signature[7]);
	printf ("%s line %i: Header size\t\t\t\t= %i\n", __FILE__, __LINE__, header.RStartPos);
	printf ("%s line %i: Header version\t\t\t\t= %i\n", __FILE__, __LINE__, header.Version);
	printf ("%s line %i: Header flags\t\t\t\t= %i\n", __FILE__, __LINE__, header.Flags);
	printf ("%s line %i: Header NumFrames\t\t\t\t= %i\n", __FILE__, __LINE__, header.NumFrames);
	printf ("%s line %i: Header width\t\t\t\t= %i\n", __FILE__, __LINE__, header.Width);
	printf ("%s line %i: Header height\t\t\t\t= %i\n", __FILE__, __LINE__, header.Height);
	printf ("%s line %i: Header block width\t\t\t= %i\n", __FILE__, __LINE__, header.BlockW);
	printf ("%s line %i: Header block height\t\t\t= %i\n", __FILE__, __LINE__, header.BlockH);
	printf ("%s line %i: Header framerate\t\t\t\t= %i\n", __FILE__, __LINE__, header.FrameRate);
	printf ("%s line %i: Header numb frames with same lookup table= %i\n", __FILE__, __LINE__, header.CBParts);
	printf ("%s line %i: Header colors\t\t\t\t= %i\n", __FILE__, __LINE__, header.Colors);
	printf ("%s line %i: Header max image blocks\t\t\t= %i\n", __FILE__, __LINE__, header.MaxBlocks);
	printf ("%s line %i: Header Unknown1\t\t\t\t= %i\n", __FILE__, __LINE__, header.Unknown1);
	printf ("%s line %i: Header Unknown2\t\t\t\t= %i\n", __FILE__, __LINE__, header.Unknown2);
	printf ("%s line %i: Header Sound freq\t\t\t= %i\n", __FILE__, __LINE__, header.Freq);
	printf ("%s line %i: Header Sound channels\t\t\t= %i\n", __FILE__, __LINE__, header.Channels);
	printf ("%s line %i: Header Sound bits\t\t\t= %i\n", __FILE__, __LINE__, header.Bits);
#endif

    // DecodeFINFChunk uses offsets
    offsets = new Uint32[header.NumFrames];
    if (!DecodeFINFChunk()) {
        delete[] offsets;
        logger->error("VQA: Invalid FINF chunk\n");
        throw runtime_error("VQA: Invalid FINF chunk\n");
    }

    CBF_LookUp = new Uint8[0x0ff00 << 3];
    CBP_LookUp = new Uint8[0x0ff00 << 3];
    VPT_Table = new Uint8[lowoffset<<1];
    CBPOffset = 0; // Starting offset of CBP Look up table must be zero
    CBPChunks = 0; // Number of CBPChunks 

    INIFile *inif = GetConfig("freecnc.ini");
    scaleVideo = inif->readInt("video", "scaleMovies", 0);
    videoScaleQuality = inif->readInt("video", "movieQuality", 0);

    //logger->debug("Video is %dfps %d %d %d\n", header.FrameRate, header.Freq, header.Channels, header.Bits);

    if (SDL_BuildAudioCVT(&cvt, AUDIO_S16SYS, header.Channels, header.Freq, SOUND_FORMAT, SOUND_CHANNELS, SOUND_FREQUENCY) < 0) {
        logger->error("Could not build SDL_BuildAudioCVT filter\n");
        return;
    }

    empty = SDL_CreateSemaphore(0);
    full = SDL_CreateSemaphore(1);
    sndBufLock = SDL_CreateMutex();

    sndbuf = new Uint8[2 * SOUND_MAX_UNCOMPRESSED_SIZE];
    sndbufMaxEnd = sndbuf + 2 * SOUND_MAX_UNCOMPRESSED_SIZE;
    sndbufStart = sndbuf;
    sndbufEnd = sndbuf;
}

VQAMovie::~VQAMovie()
{
    delete[] CBF_LookUp;
    delete[] CBP_LookUp;
    delete[] VPT_Table;
    delete[] offsets;
    VFS_Close(vqafile);

    SDL_DestroySemaphore(empty);
    SDL_DestroySemaphore(full);
    SDL_DestroyMutex(sndBufLock);

    delete[] sndbuf;
}

/// Helper function for SDL_mixer to play VQA sound and synchronise
void VQAMovie::AudioHook(void* udata, Uint8* stream, int len)
{
    VQAMovie* vqa = (VQAMovie* )udata;

    // We only need half of what they ask for, since SDL_ConvertAudio 
    // makes our audio bigger
    len = len / vqa->cvt.len_mult;

    // While the buffer isn't filled, wait
    SDL_LockMutex(vqa->sndBufLock);
    while ((vqa->sndbufEnd - vqa->sndbufStart) < len) {
        SDL_UnlockMutex(vqa->sndBufLock);
        SDL_SemWait(vqa->empty);
        SDL_LockMutex(vqa->sndBufLock);
    }

    // Copy our buffered data into the stream
    memcpy(stream, vqa->sndbufStart, len);

    // Convert the audio to the format we like
    vqa->cvt.buf = stream;
    vqa->cvt.len = len;

    if (SDL_ConvertAudio(&vqa->cvt) < 0) {
        logger->warning("Could not run conversion filter: %s\n", SDL_GetError());
        return;
    }

    // Move the start up
    vqa->sndbufStart += len;

    // If we are too near the end, move back some
    if ((vqa->sndbufEnd + SOUND_MAX_UNCOMPRESSED_SIZE) >= vqa->sndbufMaxEnd) {

        int len = (int)(vqa->sndbufEnd - vqa->sndbufStart);

        // Move remaining data down
        memmove(vqa->sndbuf, vqa->sndbufStart, len);

        // Reset the start & ends
        vqa->sndbufStart = vqa->sndbuf;
        vqa->sndbufEnd = vqa->sndbufStart + len;
    }

    SDL_SemPost(vqa->full);

    SDL_UnlockMutex(vqa->sndBufLock);
}

/// Play the vqamovie
void VQAMovie::play()
{
    if (vqafile == 0)
        return;

    if (pc::sfxeng->NoSound()) {
        return;
    }
    SDL_Surface *frame, *cframe;
    SDL_Rect dest;
    SDL_Event esc_event;
    static ImageProc scaler;

    dest.w = header.Width<<1;
    dest.h = header.Height<<1;
    dest.x = (pc::gfxeng->getWidth()-(header.Width<<1))>>1;
    dest.y = (pc::gfxeng->getHeight()-(header.Height<<1))>>1;

    pc::gfxeng->clearScreen();

    // Seek to first frame/snd information of vqa
    vqafile->seekSet(offsets[0]);

    sndindex = 0;
    sndsample = 0;

    // create the frame to store the image in. 
    frame = SDL_CreateRGBSurface(SDL_SWSURFACE, header.Width, header.Height, 8, 0, 0, 0, 0);

    // Initialise the scaler 
    if (scaleVideo)
        scaler.initVideoScale(frame, videoScaleQuality);

    Uint32 jumplen;
    Uint32 sndlen = DecodeSNDChunk(sndbufEnd);
    sndbufEnd += sndlen;
#ifdef DEBUG_VQA
	printf ("%s line %i: Decode SND done, size = %i\n", __FILE__, __LINE__, sndlen);
#endif

    // The jump len is how far ahead we are allowed to be
    jumplen = sndlen;

    // Start music (aka the sound)
    pc::sfxeng->SetMusicHook(AudioHook, this);
    for (Uint32 framenum = 0; framenum < header.NumFrames; ++framenum) {
        SDL_LockMutex(sndBufLock);

        // Decode SND Chunk first 
        sndlen = DecodeSNDChunk(sndbufEnd);
        sndbufEnd += sndlen;

        // Signal we have more audio data
        SDL_SemPost(empty);

        // Wait until the buffer has been emptied again
	if (sndlen > 0){
	        while ((sndbufEnd - sndbufStart) >= (Sint32)(jumplen - (2 * sndlen))) {
        	    SDL_UnlockMutex(sndBufLock);
	            SDL_SemWait(full);
        	    SDL_LockMutex(sndBufLock);
	        }
	}

        SDL_UnlockMutex(sndBufLock);

        if (!DecodeVQFRChunk(frame)) {
            logger->error("VQA: Decoding VQA frame\n");
            break;
        }

        if (scaleVideo) {
            cframe = scaler.scaleVideo(frame);
            pc::gfxeng->drawVQAFrame(cframe);
        } else {
            pc::gfxeng->drawVQAFrame(frame);
        }

        while (SDL_PollEvent(&esc_event)) {
            if (esc_event.type == SDL_KEYDOWN) {
                if (esc_event.key.state != SDL_PRESSED) {
                    break;
                }
                if (esc_event.key.keysym.sym == SDLK_ESCAPE) {
                    framenum = header.NumFrames; // Jump to end of video
                    break;
                }
            }
        }
    }

    pc::sfxeng->SetMusicHook(0, 0);

    SDL_FreeSurface(frame);
    if (scaleVideo) {
        scaler.closeVideoScale();
    }
}

bool VQAMovie::DecodeFORMChunk()
{
    char chunkid[4];

//	Uint16 Startpos = vqafile->tell();

//	printf ("Curpos = %i\n", vqafile->tell());

    vqafile->readByte((Uint8*)chunkid, 4);

    if (strncmp(chunkid, "FORM", 4)) {
        logger->error("VQA: Decoding FORM Chunk - Expected \"FORM\", got \"%c%c%c%c\"\n",
                chunkid[0], chunkid[1], chunkid[2], chunkid[3]);
        return false;
    }

    Uint32 chunklen;
    vqafile->readDWord(&chunklen, 1);
    chunklen = SDL_Swap32(chunklen) + 42;

//	printf ("FORM chunck Length = %i\n", chunklen);

    // skip chunklen 
//    vqafile->seekCur(4);
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
    vqafile->readByte(reinterpret_cast<Uint8 *>(&header), header_size);
#else
    vqafile->readByte((Uint8*)&header.Signature, sizeof(header.Signature));
    vqafile->readDWord(&header.RStartPos, 1);
    vqafile->readWord(&header.Version, 1);
    vqafile->readWord(&header.Flags, 1);
    vqafile->readWord(&header.NumFrames, 1);
    vqafile->readWord(&header.Width, 1);
    vqafile->readWord(&header.Height, 1);
    vqafile->readByte(&header.BlockW, 1);
    vqafile->readByte(&header.BlockH, 1);
    vqafile->readByte(&header.FrameRate, 1);
    vqafile->readByte(&header.CBParts, 1);
    vqafile->readWord(&header.Colors, 1);
    vqafile->readWord(&header.MaxBlocks, 1);
    vqafile->readWord(&header.Unknown1, 1);
    vqafile->readDWord(&header.Unknown2, 1);
    vqafile->readWord(&header.Freq, 1);
    vqafile->readByte(&header.Channels, 1);
    vqafile->readByte(&header.Bits, 1);
    vqafile->readByte((Uint8*)&header.Unknown3, sizeof(header.Unknown3));
#endif
    // Weird: need to byteswap on both BE and LE
    // readDWord probably swaps back on BE.
    header.RStartPos = SDL_Swap32(header.RStartPos);
    if (strncmp((const char*)header.Signature, "WVQAVQHD", 8) == 1) {
        logger->error("VQA: Invalid header: Expected \"WVQAVQHD\", got \"%c%c%c%c%c%c%c%c\"\n",
                header.Signature[0], header.Signature[1], header.Signature[2],
                header.Signature[3], header.Signature[4], header.Signature[5],
                header.Signature[6], header.Signature[7]);
        return false;
    }
    if (header.Version != 2) {
        logger->error("VQA: Unsupported version: Expected 2, got %i\n", header.Version);
        return false;
    }
    // Set some constants based on the header
    lowoffset = (header.Width/header.BlockW)*(header.Height/header.BlockH);
    modifier = header.BlockH == 2 ? 0x0f : 0xff;

//	vqafile->seekSet (Startpos + chunklen);
    return true;
}

/* endian macros */
#if !defined(BE_16) || !defined(BE_32) || !defined(LE_16) || !defined(LE_32)
#define BE_16(x)  ((((Uint8*)(x))[0] << 8) | ((uint8_t*)(x))[1])
#define BE_32(x)  ((((Uint8*)(x))[0] << 24) | \
                    (((Uint8*)(x))[1] << 16) | \
                    (((Uint8*)(x))[2] << 8) | \
                     ((Uint8*)(x))[3])
#define LE_16(x)  ((((Uint8*)(x))[1] << 8) | ((uint8_t*)(x))[0])
#define LE_32(x)  ((((Uint8*)(x))[3] << 24) | \
                 (((Uint8*)(x))[2] << 16) | \
                  (((Uint8*)(x))[1] << 8) | \
                    ((Uint8*)(x))[0])
#endif

bool VQAMovie::ReadChunk ()
{
	printf ("%s line %i: ***Read Chunk***\n", __FILE__, __LINE__);
	printf ("%s line %i: Chunk Start = %i\n", __FILE__, __LINE__, vqafile->tell());

	Uint8 chunkid[4];
	vqafile->readByte(chunkid, 4);

	printf ("Chunkid = %c%c%c%c\n", chunkid[0], chunkid[1], chunkid[2], chunkid[3]);

	// We have read the PINF chunk...
	Uint32 len;
	//Uint8  byte;
	vqafile->readDWord(&len, 1);
	len = BE_32 (&len);
	//len = SDL_Swap32(len);
	printf ("Length = %i\n", len);
	vqafile->seekSet(vqafile->tell() + len);

	printf ("Curpos = %i\n", vqafile->tell());

	// try again
//	vqafile->readByte(chunkid, 4);
//	printf ("%s line %i: Chunkid = %c%c%c%c\n", __FILE__, __LINE__, chunkid[0], chunkid[1], chunkid[2], chunkid[3]);
	return true;
}

bool VQAMovie::DecodeFINFChunk()
{
	Uint8 chunkid[4];
	vqafile->readByte(chunkid, 4);

	if (strncmp((const char*)chunkid, "FINF", 4) ) {
		// if this chunk id is PINF skip it..

		// Scip PINF chunk
		if (strncmp((const char*)chunkid, "PINF", 4) == 0) {
			// We have read the PINF chunk...
			Uint32 len;
			vqafile->readDWord(&len, 1);
			len = SDL_Swap32(len);
			vqafile->seekSet(vqafile->tell() + len);
			// try again
			vqafile->readByte(chunkid, 4);
		}

		if (strncmp((const char*)chunkid, "FINF", 4) ) {
			logger->error("VQA: Decoding FINF chunk - Expected \"FINF\", got \"%c%c%c%c\"\n", chunkid[0], chunkid[1], chunkid[2], chunkid[3]);
			return false;
		}
	}
//	printf ("%s line %i: Now at chunk: %s\n", __FILE__, __LINE__, chunkid);

	Uint32 chunklen;
	vqafile->readDWord(&chunklen, 1);
	chunklen = SDL_Swap32(chunklen);
	if (static_cast<Uint32>(header.NumFrames << 2) != chunklen) {
		logger->error("VQA: Invalid chunk length (%i != %i)\n", header.NumFrames << 2, chunklen);
		return false;
	}

	vqafile->readByte((Uint8*)offsets, header.NumFrames<<2);
	Uint32* offp = offsets;
	for (Uint32 framenum = 0; framenum < header.NumFrames; ++framenum, ++offp) {
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
		*offp = SDL_Swap32(*offp);
#endif
		*offp &= 0x3FFFFFFF;
		*offp <<= 1;
	}
	return true;
}

/** Decodes SND Chunk
 * @param pointer to store the decoded chunk
 * @return length of chunk
 */
Uint32 VQAMovie::DecodeSNDChunk(Uint8 *outbuf)
{
    Uint32 chunklen;
    Uint8 *inbuf;

    // header Flags tells us that this VQA does not support sounds.. so lets quit
    if (!(header.Flags & 1))
        return 0;

    // seek to correct offset
    if (vqafile->tell() & 1)
        vqafile->seekCur(1);

    Uint8 chunkid[4];

	vqafile->readByte(chunkid, 4);

	if (strncmp((const char*)chunkid, "SND", 3) ) {
		// Scip SN2J chunk's
		if (strncmp((const char*)chunkid, "SN2J", 4) == 0) {
			// We have read the PINF chunk...
			Uint32 len;
			vqafile->readDWord(&len, 1);
			len = SDL_Swap32(len);
			vqafile->seekSet(vqafile->tell() + len);
			// try again
			vqafile->readByte(chunkid, 4);
		}
		if (strncmp((const char*)chunkid, "SND", 3) ) {
			logger->error("VQA: Decoding SND chunk - Expected \"SNDX\", got \"%c%c%c%c\"\n", chunkid[0], chunkid[1], chunkid[2], chunkid[3]);
			return 0; // Returning zero here, to set length of sound chunk to zero
		}
	}

    vqafile->readDWord(&chunklen, 1);
    chunklen = SDL_Swap32(chunklen);

    inbuf = new Uint8[chunklen];

    vqafile->readByte(inbuf, chunklen);

    switch (chunkid[3]/*VQA_HI_BYTE(chunkid)*/) {
    case '0': // Raw uncompressed wave data 
        memcpy(outbuf, inbuf, chunklen);
        break;
    case '1': // Westwoods own algorithm
        // TODO: Add support for this algorithm
        logger->error("VQA: Decoding SND chunk - sound compressed using unsupported westwood algorithm\n");
        //Sound::WSADPCM_Decode(outbuf, inbuf, chunklen, uncompressed_size)
        chunklen = 0;
        break;
    case '2': // IMA ADPCM algorithm 
        Sound::IMADecode(outbuf, inbuf, chunklen, sndsample, sndindex);
        chunklen <<= 2; // IMA ADPCM decompresses sound to a size 4 times larger than the compressed size 
        break;
    default:
        logger->error("VQA: Decoding SND chunk - sound in unknown format\n");
        chunklen = 0;
        break;
    }

    delete[] inbuf;

    return chunklen;
}

/** Decodes VQFR Chunk into one frame(?)
 * @param pointer to decoded frame
 * @param Current Frame to decode
 */
bool VQAMovie::DecodeVQFRChunk(SDL_Surface *frame)
{
    Uint32 chunkid;
    Uint32 chunklen;
    Uint8 HiVal, LoVal;
    Uint8 CmpCBP, compressed; // Is CBP Look up table compressed or not
    int cpixel, bx, by, fpixel;
    SDL_Color CPL[256];

    if (vqafile->tell() & 1) {
        vqafile->seekCur(1);
    }
/*
    vqafile->readDWord(&chunkid, 1);
    if (chunkid != vqa_vqfr_id) {
        logger->error("VQA: Decoding VQFR Chunk - Expected 0x%X, got 0x%X\n",
                vqa_vqfr_id, chunkid);
        return false;
    }
*/
	Uint8 chunkid1[4];
	vqafile->readByte(chunkid1, 4);
	if (strncmp((const char*)chunkid1, "VQFR", 3) ) {
		logger->error("VQA: Decoding VQFR chunk - Expected \"VQFR\", got \"%c%c%c%c\"\n", chunkid1[0], chunkid1[1], chunkid1[2], chunkid1[3]);
		return 0; // Returning zero here, to set length of sound chunk to zero
	}

    vqafile->readDWord(&chunklen, 1);
    chunklen = SDL_Swap32(chunklen);
    CmpCBP = 0;

    bool done = false;
    // Read chunks until we get to the VPT chunk
    while (!done) {
        if (vqafile->tell() & 1)
            vqafile->seekCur(1);

        vqafile->readDWord(&chunkid, 1);
        compressed = VQA_HI_BYTE(chunkid) == 'Z';
        chunkid = chunkid & vqa_t_mask;
        switch (chunkid) {
        case vqa_cpl_id:
            DecodeCPLChunk(CPL);
            SDL_SetColors(frame, CPL, 0, header.Colors);
            break;
        case vqa_cbf_id:
            DecodeCBFChunk(compressed);
            break;
        case vqa_cbp_id:
            CmpCBP = compressed;
            DecodeCBPChunk();
            break;
        case vqa_vpt_id:
            // This chunk is always the last one 
            DecodeVPTChunk(compressed);
            done = true;
            break;
        default:
            DecodeUnknownChunk();
        }
    }

    // Optimise me harder
    cpixel = 0;
    fpixel = 0;
    for (by = 0; by < header.Height; by += header.BlockH) {
        for (bx = 0; bx < header.Width; bx += header.BlockW) {
            LoVal = VPT_Table[fpixel]; // formerly known as TopVal 
            HiVal = VPT_Table[fpixel + lowoffset]; // formerly known as LowVal 
            if (HiVal == modifier) {
                memset((Uint8 *)frame->pixels + cpixel, LoVal, header.BlockW);
                memset((Uint8 *)frame->pixels + cpixel + header.Width, LoVal, header.BlockW);
            } else {
                memcpy((Uint8 *)frame->pixels + cpixel, &CBF_LookUp[((HiVal<<8)|LoVal)<<3], header.BlockW);
                memcpy((Uint8 *)frame->pixels + cpixel + header.Width,
                       &CBF_LookUp[(((HiVal<<8)|LoVal)<<3)+4], header.BlockW);
            }
            cpixel += header.BlockW;
            fpixel++;
        }
        cpixel += header.Width;
    }

    //    if( !((CurFrame + 1) % header.CBParts) ) {
    if (CBPChunks & ~7) {
        if (CmpCBP == 1) {
            Uint8 CBPUNZ[0x0ff00 << 3];
            Compression::decode80(CBP_LookUp, CBPUNZ);
            memcpy(CBF_LookUp, CBPUNZ, 0x0ff00 << 3);
        } else {
            memcpy(CBF_LookUp, CBP_LookUp, 0x0ff00 << 3);
        }
        CBPOffset = 0;
        CBPChunks = 0;
    }
    return true;
}

inline void VQAMovie::DecodeCBPChunk()
{
    Uint32 chunklen;

    vqafile->readDWord(&chunklen, 1);
    chunklen = SDL_Swap32(chunklen);

    vqafile->readByte(CBP_LookUp + CBPOffset, chunklen);
    CBPOffset += chunklen;
    CBPChunks++;
}

inline void VQAMovie::DecodeVPTChunk(Uint8 Compressed)
{
    Uint32 chunklen;

    vqafile->readDWord( &chunklen, 1);
    chunklen = SDL_Swap32(chunklen);

    if (Compressed) {
        Uint8 *VPTZ; // Compressed VPT_Table 
        VPTZ = new Uint8[chunklen];
        vqafile->readByte(VPTZ, chunklen);
        Compression::decode80(VPTZ, VPT_Table);
        delete[] VPTZ;
    } else { // uncompressed VPT chunk. never found any.. but might be some 
        vqafile->readByte(VPT_Table, chunklen);
    }
}

inline void VQAMovie::DecodeCBFChunk(Uint8 Compressed)
{
    Uint32 chunklen;

    vqafile->readDWord(&chunklen, 1);
    chunklen = SDL_Swap32(chunklen);

    if (Compressed) {
        Uint8 *CBFZ; // Compressed CBF table 
        CBFZ = new Uint8[chunklen];
        vqafile->readByte(CBFZ, chunklen);
        Compression::decode80(CBFZ, CBF_LookUp);
        delete[] CBFZ;
    } else {
        vqafile->readByte(CBF_LookUp, chunklen);
    }
}

inline void VQAMovie::DecodeCPLChunk(SDL_Color *palette)
{
    Uint32 chunklen;
    int i;

    vqafile->readDWord(&chunklen, 1);
    chunklen = SDL_Swap32(chunklen);

    for (i = 0; i < header.Colors; i++) {
        vqafile->readByte((Uint8*)&palette[i], 3);
        palette[i].r <<= 2;
        palette[i].g <<= 2;
        palette[i].b <<= 2;
    }
}

inline void VQAMovie::DecodeUnknownChunk()
{
    Uint32 chunklen;

    vqafile->readDWord(&chunklen, 1);
    chunklen = SDL_Swap32(chunklen);
    logger->debug("Unknown chunk at %0x for %i bytes.\n", vqafile->tell(), chunklen);

    vqafile->seekCur(chunklen);
}
